// G15_Display - DLL to interface with MUSHclient via Lua to display information
//    on the Logitech G15 gamers keyboard.
//  Author: Nick Gammon
//  Date:   16th June 2006
//  Version: 1.1
//


/*
Copyright  2006 Nick Gammon.

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the "Software"), to deal 
in the Software without restriction, including without limitation the rights 
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.    

The software is provided "as is", without warranty of any kind, express or 
implied, including but not limited to the warranties of merchantability, 
fitness for a particular purpose and noninfringement. In no event shall the 
authors or copyright holders be liable for any claim, damages or other 
liability, whether in an action of contract, tort or otherwise, arising from, 
out of or in connection with the software or the use or other dealings in the 
software. 
*/

#include "EZ_LCD.h"

#define VERSION "1.1"

#ifdef _WIN32
#define LUA_API __declspec(dllexport)
#endif

#pragma comment( lib, "lua50.lib" )

extern "C" {

#include "lua.h"

#include "lauxlib.h"
#include "lualib.h"

  }

CEzLcd* lcd = NULL;   // pointer to LCD stuff

const char text_handle[] = "g15_text_handle";
const char progress_bar_handle[] = "g15_progress_bar_handle";

//---------------------------------------------------------------------
//  general checks
//---------------------------------------------------------------------

static void check_opened (lua_State *L)
  {
  if (!lcd)
    luaL_error (L, "LCD device not opened");
  }  // end of check_opened

static void check_connected (lua_State *L)
  {
  check_opened (L);

  // maybe we won't care

//  if (!lcd->IsConnected())
//    luaL_error (L, "LCD device not connected");
  }  // end of check_connected

static void mushclient_debug (lua_State *L, const char * message)
  {
  lua_getglobal (L, "world");   /* world functions */
  lua_pushliteral (L, "ColourNote"); /* ColourNote function */
  lua_gettable (L, -2);  /* get function */
  lua_pushliteral (L, "white");  /* foreground colour */
  lua_pushliteral (L, "blue");   /* background colour */
  lua_pushstring  (L, message);  /* message */
  lua_call (L, 3, 0);  /* call with 3 args and no result */
  }

//---------------------------------------------------------------------
//  for testing a boolean
//---------------------------------------------------------------------

static BOOL get_boolean (lua_State *L, const int narg)
  {

  if (lua_isnil (L, narg))
    return FALSE;

  // non-nil is true unless the exact value: boolean (false)
  if (lua_isboolean (L, narg))
    return lua_toboolean (L, narg);

  return TRUE;
  } // end of get_boolean

//---------------------------------------------------------------------
//  TEXT functions (for userdata created with AddText)
//---------------------------------------------------------------------

// check our userdata (self parameter) is what we expect
static HANDLE check_text_handle (lua_State *L)
  {
  check_opened (L);

  HANDLE * ud = (HANDLE *) luaL_checkudata(L, 1, text_handle);
  luaL_argcheck(L, *ud != NULL, 1, "lcd text type expected");
  return *ud;
  }  // end of check_text_handle

// text_handle:set (contents)
//  eg.  th = g15.AddText ("static", "small", "left", 160, 0, 0)
//       th:set ("score = 15")
static int text_set (lua_State *L)
  {
  HANDLE h = check_text_handle (L);

  check_connected (L);

  const char * s = luaL_checkstring (L, 2);  // arg 1 is self
	lcd->SetText(h, s);
	lcd->Update();
  return 0;
  } // end of text_set

// make text visible or not
static int text_visible (lua_State *L)
  {
  HANDLE h = check_text_handle (L);
  check_connected (L);

  lcd->SetVisible (h, get_boolean (L, 2));
	lcd->Update();
  return 0;
  } // end of text_visible

// garbage-collect by deleting text handle
static int text_gc (lua_State *L)
  {
  HANDLE h = check_text_handle (L);
//  delete h;      // seems to crash it ;)
//  mushclient_debug (L, "text handle garbage collection");
  return 0;
  } // end of text_gc

// for displaying what userdata is
static int text_tostring (lua_State *L)
  {
  HANDLE h = check_text_handle (L);
  lua_pushstring (L, "lcd text display");
  return 1;
  } // end of text_tostring

// text methods
static const luaL_reg textmeta[] = {
  {"set",        text_set},
  {"visible",    text_visible},
  {"__gc",       text_gc},
  {"__tostring", text_tostring},
  {NULL, NULL}
};   // end of textmeta


//---------------------------------------------------------------------
//  PROGRESS BAR functions (for userdata created with AddProgressBar)
//---------------------------------------------------------------------

// check our userdata (self parameter) is what we expect
static HANDLE check_progress_bar_handle (lua_State *L)
  {
  check_opened (L);

  HANDLE * ud = (HANDLE *) luaL_checkudata(L, 1, progress_bar_handle);
  luaL_argcheck(L, *ud != NULL, 1, "lcd progress bar type expected");
  return *ud;
  }  // end of check_progress_bar_handle

// progress_bar_handle:set (contents)
//  eg.  ph = g15.AddProgressBar ("filled", 100, 5, 0, 0)
//       ph:percent (42)
static int progress_bar_percent (lua_State *L)
  {
  HANDLE h = check_progress_bar_handle (L);

  check_connected (L);

  int iPercent = (int) luaL_checknumber (L, 2);  // arg 1 is self
  if (iPercent > 100)    // force into range 0 to 100
    iPercent = 100;
  else if (iPercent < 0)
    iPercent = 0;
	lcd->SetProgressBarPosition(h, iPercent);
	lcd->Update();
  return 0;
  } // end of progress_bar_percent

// make progress bar visible or not
static int progress_bar_visible (lua_State *L)
  {
  HANDLE h = check_progress_bar_handle (L);
  check_connected (L);

  lcd->SetVisible (h, get_boolean (L, 2));
	lcd->Update();
  return 0;
  } // end of progress_bar_visible

// garbage-collect by deleting text handle
static int progress_bar_gc (lua_State *L)
  {
  HANDLE h = check_progress_bar_handle (L);
//  delete h;      // seems to crash it ;)
//  mushclient_debug (L, "progress bar garbage collection");
  return 0;
  } // end of progress_bar_gc

// for displaying what userdata is
static int progress_bar_tostring (lua_State *L)
  {
  HANDLE h = check_progress_bar_handle (L);
  lua_pushstring (L, "lcd progress bar display");
  return 1;
  } // end of progress_bar_tostring

// text methods
static const luaL_reg progress_barmeta[] = {
  {"percent",    progress_bar_percent},
  {"visible",    progress_bar_visible},
  {"__gc",       progress_bar_gc},
  {"__tostring", progress_bar_tostring},
  {NULL, NULL}
};   // end of progress_barmeta


//---------------------------------------------------------------------
//  general functions (in the g15 table)
//---------------------------------------------------------------------

// see if device connected
static int IsConnected (lua_State *L)
  {
  check_opened (L);
  lua_pushboolean (L, lcd->IsConnected());
  return 1;
  }  // end of IsConnected

// see if we opened it OK
static int IsOpen (lua_State *L)
  {
  lua_pushboolean (L, lcd != NULL);
  return 1;
  } // end of IsOpen

// force an update
static int Update (lua_State *L)
  {
  check_connected (L);

	lcd->Update();
  return 0;
  }   // end of Update

// force an update
static int SetAsForeground (lua_State *L)
  {
  check_connected (L);
	lcd->SetAsForeground (get_boolean (L, 2));
  return 0;
  }   // end of SetAsForeground

static LGSoftButton get_button_number  (lua_State *L)
  {
  switch ((int) luaL_checknumber (L, 1))
    {
    case 1: return LG_BUTTON_1;
    case 2: return LG_BUTTON_2;
    case 3: return LG_BUTTON_3;
    case 4: return LG_BUTTON_4;
    } // end of switch

  luaL_error (L, "button number must be 1 to 4");
  return LG_BUTTON_1;  // not reached
  }

// Check if a button was triggered
static int ButtonTriggered (lua_State *L)
  {
  check_connected (L);
  lua_pushboolean (L, lcd->ButtonTriggered (get_button_number (L)));
  return 1;
  }   // end of ButtonTriggered

// Check if a button was released
static int ButtonReleased (lua_State *L)
  {
  check_connected (L);
  lua_pushboolean (L, lcd->ButtonReleased (get_button_number (L)));
  return 1;
  }   // end of ButtonReleased

// Check if a button is being pressed right now
static int ButtonIsPressed (lua_State *L)
  {
  check_connected (L);
  lua_pushboolean (L, lcd->ButtonIsPressed (get_button_number (L)));
  return 1;
  }   // end of ButtonIsPressed

// g15.addtext (type, size, align, pixels, x, y)

//  type = one of:  static/scrolling
//  size = one of:  small/medium/big
//  align = one of: left/center/right
//  pixels = maximum width in pixels (eg. up to 160)
//  x = X-position (0 = left)
//  y = Y-position (0 = top)

static int AddText (lua_State *L)
  {
  check_opened (L);

  const char * sType = luaL_optstring (L, 1, "static");

  LGObjectType iType;

  if (strcmp (sType, "static") == 0)
    iType = LG_STATIC_TEXT;
  else if (strcmp (sType, "scrolling") == 0) 
    iType = LG_SCROLLING_TEXT;
  else
    luaL_error (L, "text type must be one of: static/scrolling");

  const char * sSize = luaL_optstring (L, 2, "small");

  LGTextSize iSize;

  if (strcmp (sSize, "small") == 0)
    iSize = LG_SMALL;
  else if (strcmp (sSize, "medium") == 0) 
    iSize = LG_MEDIUM;
  else if (strcmp (sSize, "big") == 0) 
    iSize = LG_BIG;
  else
    luaL_error (L, "text size must be one of: small/medium/big");

  const char * sAlign = luaL_optstring (L, 3, "left");

  INT iAlign;

  if (strcmp (sAlign, "left") == 0)
    iAlign = DT_LEFT;
  else if (strcmp (sAlign, "center") == 0) 
    iAlign = DT_CENTER;
  else if (strcmp (sAlign, "right") == 0) 
    iAlign = DT_RIGHT;
  else
    luaL_error (L, "text alignment must be one of: left/center/right");

  const int iMax = (int) luaL_optnumber (L, 4, 80);

  if (iMax < 1)
    luaL_error (L, "max pixels must be >= 1");

  const int iLeft = (int) luaL_optnumber (L, 5, 0);

  if (iLeft < 0)
    luaL_error (L, "origin (X) must be >= 0");

  const int iTop = (int) luaL_optnumber (L, 6, 0);

  if (iTop < 0)
    luaL_error (L, "origin (Y) must be >= 0");

  // make our userdata (which is a pointer to a handle)
  HANDLE * ud = (HANDLE *) lua_newuserdata (L, sizeof(HANDLE));
  luaL_getmetatable(L, text_handle);
  lua_setmetatable(L, -2);

 	*ud = lcd->AddText(iType, iSize, iAlign, iMax);
	lcd->SetOrigin(*ud, iLeft, iTop);
	lcd->Update();

  return 1;   // the new handle
  } /* end of AddText */


// g15.AddProgressBar (type, sizeX, sizeY, x, y)

//  type = one of:  cursor/filled/dot
//  sizeX = width in pixels
//  sizeY = height in pixels
//  x = X-position (0 = left)
//  y = Y-position (0 = top)

static int AddProgressBar (lua_State *L)
  {
  check_opened (L);

  const char * sType = luaL_optstring (L, 1, "cursor");

  LGProgressBarType iType;

  if (strcmp (sType, "cursor") == 0)
    iType = LG_CURSOR;
  else if (strcmp (sType, "filled") == 0) 
    iType = LG_FILLED;
  else if (strcmp (sType, "dot") == 0) 
    iType = LG_DOT_CURSOR;
  else
    luaL_error (L, "text type must be one of: cursor/filled/dot");


  const int iSizeX = (int) luaL_optnumber (L, 2, 120);

  if (iSizeX < 1)
    luaL_error (L, "size X must be >= 1");

  const int iSizeY = (int) luaL_optnumber (L, 3, 5);

  if (iSizeY < 1)
    luaL_error (L, "size Y must be >= 1");

  const int iLeft = (int) luaL_optnumber (L, 4, 0);

  if (iLeft < 0)
    luaL_error (L, "origin (X) must be >= 0");

  const int iTop = (int) luaL_optnumber (L, 5, 0);

  if (iTop < 0)
    luaL_error (L, "origin (Y) must be >= 0");

  // make our userdata (which is a pointer to a handle)
  HANDLE * ud = (HANDLE *) lua_newuserdata (L, sizeof(HANDLE));
  luaL_getmetatable(L, progress_bar_handle);
  lua_setmetatable(L, -2);

 	*ud = lcd->AddProgressBar(iType);
	lcd->SetProgressBarSize(*ud, iSizeX, iSizeY);
	lcd->SetOrigin(*ud, iLeft, iTop);
	lcd->Update();

  return 1;   // the new handle
  } /* end of AddProgressBar */


// helper function to create metatables
static void createmeta(lua_State *L, const char *name)
{
  luaL_newmetatable(L, name);   /* create new metatable */
  lua_pushliteral(L, "__index");
  lua_pushvalue(L, -2);         /* push metatable */
  lua_rawset(L, -3);            /* metatable.__index = metatable */
}

// functions supported by the g15 library
static const luaL_reg g15lib[] = 
{
  {"IsOpen", IsOpen},
  {"IsConnected", IsConnected},
  {"Update", Update},
  {"AddText", AddText},
  {"AddProgressBar", AddProgressBar},
  {"SetAsForeground", SetAsForeground},
  {"ButtonTriggered", ButtonTriggered},
  {"ButtonReleased",  ButtonReleased},
  {"ButtonIsPressed", ButtonIsPressed},
  {NULL, NULL}
};

extern "C" {

/*
** Open G15 library
*/
LUALIB_API int luaopen_g15 (lua_State *L)
 {
  // install g15 library into Lua address space
  luaL_openlib(L, "g15", g15lib, 0);

  // metatable for text objects
  createmeta(L, text_handle);
  luaL_openlib(L, NULL, textmeta, 0);
  lua_pop(L, 1);    // pop table it was put into

  // metatable for progress bar objects
  createmeta(L, progress_bar_handle);
  luaL_openlib(L, NULL, progress_barmeta, 0);
  lua_pop(L, 1);    // pop table it was put into


  HRESULT hRes;

  // Create instance of EzLcd.
  lcd = new CEzLcd();

  // Have it initialize itself
  hRes = lcd->InitYourself(_T("MUSHclient"));

  if (hRes != S_OK)
    {
    // Something went wrong, when connecting to the LCD Manager software. 
    delete lcd;
    lcd = NULL;
    }

  if (lcd)
    {
	  lcd->Update();
    mushclient_debug (L, "g15 keyboard LCD display DLL loaded - Version " VERSION);
    }


  return 1;
 }   // end of luaopen_g15

}  // end of: extern "C"
